{-# LANGUAGE NamedFieldPuns #-}
{-# LANGUAGE PatternGuards #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE BangPatterns #-}
module Engine.RuntimePlan
( numberOfLeaves
, PlanStrategy(..)
, updateRuntimePlan
, initRuntimePlan
, RuntimePlan
, dumpRuntimePlan
) where

import Control.Applicative
import Control.Arrow
import Data.Monoid
import Data.Sequence (Seq, ViewL((:<)), (<|), (><))
import Engine.EngineM
import Engine.Experiment
import Engine.ExperimentResult
import Plan
import qualified Data.Foldable as F
import qualified Data.Sequence as Seq
import Util

data RuntimePlan a
  = Unknown !a
  | Split { children        :: !(Seq (RuntimePlan a))
          , _numberOfLeaves :: !Int
          }

dumpRuntimePlan :: RuntimePlan a -> String
dumpRuntimePlan (Unknown a) = "."
dumpRuntimePlan (Split { children }) = "(" ++ concat (F.toList $ (dumpRuntimePlan <$> children)) ++ ")"

class PlanStrategy a where
  select :: Seq (RuntimePlan a) -> EngineM Int
  update :: a -> EngineM (Seq a, Maybe (Experiment, RanExperiment))

initRuntimePlan :: a -> RuntimePlan a
initRuntimePlan = Unknown

numberOfLeaves :: RuntimePlan a -> Int
numberOfLeaves (Unknown _)                     = 1
numberOfLeaves (Split { _numberOfLeaves = x }) = x


updateRuntimePlan :: PlanStrategy a => RuntimePlan a -> EngineM (Maybe (RuntimePlan a), (Maybe (Experiment, RanExperiment)))
updateRuntimePlan (Unknown a) = profileFunction "updateRuntimePlan" $ first mkResult <$> update a
  where
    mkResult :: Seq a -> Maybe (RuntimePlan a)
    mkResult (Seq.viewl -> Seq.EmptyL)        = Nothing
    mkResult (Seq.viewl -> (a :< as))
      | Seq.null as                   = Just (Unknown a)
    mkResult s                        = Just (Split { children = Unknown <$> s
                                                    , _numberOfLeaves = Seq.length s
                                                    })

updateRuntimePlan (Split { children }) 
  | Seq.null children = profileFunction "updateRuntimePlan" $ do
    return (Nothing, Nothing)
  | otherwise         = profileFunction "updateRuntimePlan" $ do
  chosenIndex <- select children
  let (before, Seq.viewl -> a :< after) = Seq.splitAt chosenIndex children

  (!rp, !b) <- updateRuntimePlan a

  !rp' <- return $! composeRuntimePlans before rp after 
  return $ (Just rp', b)

    
composeRuntimePlans :: Seq (RuntimePlan a) -> Maybe (RuntimePlan a) -> Seq (RuntimePlan a) -> (RuntimePlan a)
composeRuntimePlans !before Nothing !after
  | numKids == 1
  , a :< _ <- Seq.viewl kids = a
  | otherwise                = Split { children = kids
                                     , _numberOfLeaves = numLeaves
                                     }
  where
    kids      = before >< after
    numKids   = Seq.length kids
    numLeaves = getSum . F.foldMap (Sum . numberOfLeaves) $ kids

composeRuntimePlans before (Just !a) after 
  = Split { children = kids
          , _numberOfLeaves = numLeaves
          }
  where
    kids = before >< (a <| after)
    numLeaves = getSum . F.foldMap (Sum . numberOfLeaves) $ kids



