{-# LANGUAGE NamedFieldPuns #-}
{-# LANGUAGE PatternGuards #-}
{-# LANGUAGE GADTs #-}
{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE MagicHash #-}
{-# LANGUAGE ForeignFunctionInterface #-}
{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# OPTIONS_GHC -fno-warn-overlapping-patterns #-}
module FrontEnd.Source
(sourceAction
) where

import Bottom
import Control.Applicative
import Control.Monad
import Control.Monad.Identity
import DataCon
import Data.List
import Data.Maybe
import Data.Monoid
import Data.Sequence (Seq, (><), (<|), (|>))
import Data.Time.Clock
import Digraph
import Distribution.Simple.Setup
import DynFlags
import Encoding
import Engine
import Foreign
import Foreign.C.String
import Foreign.C.Types
import Format
import FrontEnd.Constants
import FrontEnd.Options
import FrontEnd.Stats
import FrontEnd.TestSuite
import GHC
import GHC.Paths
import HscTypes
import IdDB
import Linker
import ObjLink
import Outputable
import Plan as Plan
import PrelNames (unpackCStringName, gHC_SHOW)
import qualified Data.Foldable as F
import qualified Data.Sequence as Seq
import qualified Util.BiMap as BiMap
import System.FilePath
import System.IO
import Trace.Hpc.Mix
import Util
import Util.BiMap (BiMap)
import Util.GHC

foreign import ccall unsafe hs_hpc_module :: Ptr CChar -> Word32 -> Word32 -> Ptr Word64 -> IO ()


data PlanStrategyRep a where
  PSR :: (PlanStrategy a) => PlanStrategyRep a

sourceAction :: [String] -> [String] -> IrulanOptions -> IO ()
sourceAction ghcOpts targets options = do
  let args = a' ++ as
        where
          as = map (mkGeneralLocated "on the commandline") (concatMap words ghcOpts)
          a' = flagToList (const (mkGeneralLocated "on the commandline" "-fhpc") <$> outputStatistics options)

  (argsStatic, warnsStatic) <- parseStaticFlags args 

  runGhc (Just libdir) $ do
    (dflags', _, warns) <- flip parseDynamicFlags argsStatic =<< getSessionDynFlags 
    let dflags'' = dflags' { hscTarget = HscAsm 
                           , ghcLink   = LinkBinary
                           }
    primPackages <- setSessionDynFlags dflags''
    dflags <- getSessionDynFlags
    -- TODO change from this to one that says 'irulan blew up' or something
    defaultErrorHandler dflags $ do
      defaultCleanupHandler dflags $ do
          let dropHsExtension = (\x -> let h = reverse x in if take 3 h == "sh." || take 4 h == "shl." 
                                                             then dropExtension x
                                                             else x)
          let userSpecifiedModules = map dropHsExtension targets
          let userSpecifiedExplicitSources = map dropHsExtension (explicitIdSource options)

          liftIO $ mapM_ (print . unLoc) (warns ++ warnsStatic)

          targets <- mapM (flip guessTarget Nothing) (nub $ userSpecifiedModules ++ userSpecifiedExplicitSources)
          setTargets targets 

          result <- load LoadAllTargets

          liftIO $ initDynLinker dflags
          do
            dflags <- getSessionDynFlags
            liftIO $ linkPackages dflags primPackages
            -- For some reason, unpackCStringName# doesn't always get loaded.
            -- Force it to be so here.
            withSession $ \hsc -> (liftIO $ Linker.getHValue hsc unpackCStringName >> return ())
            _ <- getModuleInfo gHC_SHOW
            return ()


          if (succeeded result) 
            then do
              mg <- depanal [] False
              let mods = flattenSCCs (topSortModuleGraph False mg Nothing)

              explicitIdentSourceMods <- mapM (flip lookupModule Nothing . mkModuleName) userSpecifiedExplicitSources

              forM_ mods $ \modsum -> do
                printWarnings

                modIface <- getModIface modsum
                
                let (modName :: String) = moduleNameString . ms_mod_name $ modsum

                linkModuleAndDependencies modsum modIface
                
                liftIO $ processHpc modName

                when (modName `elem` userSpecifiedModules) $ do
                  liftIO $ putStrLn (modName ++ ":")

                  constants <- makeConstants (intConstants options)
                                             (charConstants options)
                                             (floatConstants options)
                                             (doubleConstants options)
                                             (integerConstants options)
                  theDB <- mkIdDBForModule (fromFlag $ useVCO options)
                                           (ms_mod modsum) 
                                           (fromFlag $ onlyTestProperties options) 
                                           (fromFlag $ sourceIdsFrom options) 
                                           explicitIdentSourceMods constants

                  if (fromFlag $ onlyDumpSupportSetSize options) 
                    then do
                      liftIO $ putStrLn (IdDB.summarise theDB)
                    else do
                      testFunctions options theDB 
                      liftIO $ processHpc modName

                `gcatch` \(e :: SourceError) -> do
                           printExceptionAndWarnings e

            else do
              liftIO $ putStrLn "No!"

processHpc :: String -> IO ()
processHpc  modName = do
  modNamePtr <- newCString modName
  mTickboxesPtr <- lookupSymbol ("_hpc_tickboxes_" ++ zEncodeString modName ++ "_hpc")
  case mTickboxesPtr of
    Nothing -> return ()
    Just tickboxesPtr -> liftIO $ do
      (Mix _ _ infoHash _ (length -> infoTickCount)) <- readMix [".hpc"] (Left modName)
      let len = fromIntegral infoTickCount
      hs_hpc_module modNamePtr (len) (read . show $ infoHash) tickboxesPtr

linkModuleAndDependencies :: ModSummary -> ModIface -> Ghc ()
linkModuleAndDependencies modsum modiface = do
    dflags <- getSessionDynFlags
    liftIO $ linkPackages dflags (dep_pkgs . mi_deps $ modiface)
    let obj_file = ml_obj_file . ms_location $ modsum
    liftIO $ loadObj obj_file
    liftIO $ (resolveObjs >> return ())
    return ()

{-# NOINLINE mkTreePlan #-}
mkTreePlan :: IdDB -> Bool -> Bool -> () -> TreePlan
mkTreePlan iddb useCaseArgs useCases () = generateTreePlan iddb useCaseArgs useCases

testFunctions :: IrulanOptions -> IdDB -> Ghc ()
testFunctions opts iddb = do
  let useCases    = fromFlag $ useCaseStatement opts
  let useCaseArgs = fromFlag $ useCaseArgument opts
  let maxDepth    = fromFlag $ maximumDepth opts
  let maxExpSize  = fromJust . getLast $ maximumExpressionSize opts
  let maxExpDepth = fromJust . getLast $ maximumExpressionDepth opts
  let treePlanF   = mkTreePlan iddb useCaseArgs useCases   
  let doID        = (fromFlag $ runtimePlanType opts) == IterativeDeepening

  when (fromFlag $ dumpPlan opts) $ do
    liftIO $ writeFile "irulan.dot" (dotifyTreePlan (fromEnum maxDepth) (treePlanF ()))

  endTime <- do timeNow <- liftIO $ getCurrentTime
                let maxRuntime = (fromFlag $ maximumRuntime opts)
                let et = Just ((fromRational . toRational $ maxRuntime) `addUTCTime` timeNow)
                case (fromFlag $ runtimePlanType opts) of
                  Random             -> return et
                  DFSEvenTime        -> return et
                  RestartingRandom   -> return et
                  DFS                -> return Nothing
                  IterativeDeepening -> return et

  timeProfile <- if (fromFlag $ doProfiling opts)
                  then liftIO emptyProfile
                  else return nullProfile

  let mkEngine maxExpDepth = case ((case (fromFlag $ runtimePlanType opts) of
                                             DFS                -> ExistsC PSR (Identity (dfss ()))
                                             IterativeDeepening -> ExistsC PSR (Identity (dfss ()))
                                             DFSEvenTime        -> ExistsC PSR (Identity (DFSET (fromJust endTime) (dfss ())))
                                             Random             -> ExistsC PSR (Identity (DLRF (dfss ())))
                                             RestartingRandom   -> ExistsC PSR (Identity (RR treePlanF maxDepth (dfss ())))
                                             ) :: ExistsClass Identity PlanStrategyRep) of
                                    ExistsC PSR (Identity plan) -> ExistsC PSR $ buildEngine 
                                                                        (fromFlag $ useShowOnResult opts)
                                                                        timeProfile
                                                                        (fromJust . getLast $ cache opts)
                                                                        (fromFlag $ maxExpressionTime opts) 
                                                                        (fromFlag $ maxExpressionAllocs opts) plan
                                  where
                                    {-# NOINLINE dfss #-}
                                    dfss x = DFSS (treePlanF x) (if doID then fromIntegral (3 * fromJust maxExpDepth) else maxDepth) maxExpDepth maxExpSize


  unless (fromFlag $ outputExperimentTrace opts) (liftIO $ putStr " ")

  liftIO $ putStrLn "Results:"
  ((), ces) <- withStatsOutput (flagToMaybe $ outputStatistics opts) (fromFlag $ outputFullStatistics opts) $ \statsOutputter -> do
                withTestSuite (fromFlag $ generateTestSuite opts) $ \testSuite -> do

                  engineLoop opts endTime (if doID then 
                                            let loop n = ({- TODO: setIterativeDeepeningDepth statsOutputter n >> -} (mkEngine $! (Just n))) : loop (n + 1) in loop 1 
                                           else [mkEngine maxExpDepth]
                                          ) 
                                          statsOutputter testSuite

                  when (fromFlag $ doProfiling opts) $ liftIO $ do
                    putStrLn "" 
                    formatProfile timeProfile >>= putStrLn

  when ((fromFlag $ generateTestSuite opts) == HpcTestSuite) $ do
    liftIO $ putStrLn "" 
    liftIO $ putStrLn "Test Suite:"
    liftIO $ F.forM_ ces $ \ce -> printError opts (ceExperiment ce) (ceResult ce) {- >> print (ceTix ce) -}

engineLoop :: IrulanOptions  -> Maybe UTCTime -> [ExistsClass Engine PlanStrategyRep] -> StatsOutputter -> TestSuite -> Ghc ()
engineLoop !opts !endTime ((!ExistsC PSR engine):engines) !statsOutputter !testSuite = do
  timeNow <- liftIO $ getCurrentTime

  if (isJust endTime) && (timeNow > fromJust endTime)
    then return ()
    else do
      let fmtFn
            | fromFlag $ useSimpleOutput opts = simpleOutput
            | otherwise                       = \exp ran -> let arr = case ran of
                                                                        NewResult _ _    -> " ==> "
                                                                        CachedResult _ _ -> " ##> "
                                                              in format exp ++ arr ++ format (getExperimentResult ran)

      stepResult <- stepEngine engine

      case stepResult of
        NoWork -> engineLoop opts endTime engines statsOutputter testSuite
        Work { experimentRun = experiment
             , ranExperiment = ranExperiment
             , nextEngine    = mNextEngine
             , debugLog      = debugLog
             } -> do
          when (fromFlag $ outputDebugInformation opts) $ 
            F.mapM_ (liftIO . putStrLn) debugLog

          if (fromFlag $ outputExperimentTrace opts)
            then (liftIO . putStrLn) (fmtFn experiment ranExperiment)
            else liftIO $ case ranExperiment of
              NewResult experimentResult@(Errors _) _       -> printError opts experiment experimentResult
              NewResult experimentResult@(Property False) _ -> printError opts experiment experimentResult
              _                                             -> return ()
              
              
          liftIO $ outputStats statsOutputter experiment ranExperiment timeNow
          liftIO $ testSuiteExperiment testSuite experiment ranExperiment
          case mNextEngine of
            Nothing          -> engineLoop opts endTime engines statsOutputter testSuite
            Just !nextEngine -> engineLoop opts endTime (ExistsC PSR nextEngine:engines) statsOutputter testSuite

engineLoop _ _ _ _ _ = return ()

printError :: IrulanOptions -> Experiment -> ExperimentResult -> IO ()
printError options experiment experimentResult = do
  if (fromFlag $ useSimpleOutput options) 
    then putStrLn ("  " ++ simpleOutput experiment (NewResult experimentResult Nothing))
    else putStrLn ("  " ++ format experiment ++ " ==> " ++ format experimentResult)

simpleOutput :: Experiment -> RanExperiment -> String
simpleOutput exp' ranExp = simpleOutput' exp ++ arrow ++ simpleOutputResult res
  where
    (arrow, res')
      | NewResult res' _    <- ranExp = (" ==> ", res')
      | CachedResult res' _ <- ranExp = (" ##> ", res')

    (exp, bimap) = canonicaliseExperiment exp'
    res = remapExperimentResult (BiMap.forwardMap bimap) res'

    simpleOutput' exp 
      = unwords $ [simpleOutputRoot (not . Seq.null $ experimentArgs exp) (experimentRoot exp)] ++ (catMaybes . F.toList $ simpleOutputArg <$> experimentArgs exp) 

    simpleOutputArg (ArgExperiment e)
      | DestructRoot {} <- experimentRoot e = Just $ "(" ++ simpleOutput' e ++ ")"
      | Seq.null (experimentArgs e)         = Just $ simpleOutput' e
      | otherwise                           = Just $ "(" ++ simpleOutput' e ++ ")"
    simpleOutputArg (ArgBottom _ b)  = Just $ simpleBottom b
    simpleOutputArg (ArgType _)      = Nothing
    simpleOutputArg ArgUnconstrained = Just $ "*"

    simpleOutputRoot _ (ConstantRoot x) = format x
    simpleOutputRoot _ (IdRoot x) = showSDocUnqual . ppr $ x
    simpleOutputRoot b (DestructRoot { destructCaseOf   
                                   , destructConstructor
                                   , destructVarIndex }) 
      | b = "(" ++ mainBit ++ ")"
      | otherwise = mainBit
      where
        mainBit = "case " ++ (simpleOutput' destructCaseOf) ++ " of " ++ dcBit ++ " -> x"
        dcBit
          | dataConIsInfix destructConstructor = head args ++ " " ++ dcString ++ " " ++ unwords (tail args)
          | otherwise = dcString ++ " " ++ unwords args
          where
            dcString = showSDocUnqual . ppr $ destructConstructor
            args = [ if y == destructVarIndex then "x" else "_" | y <- [0 ..(length $ dataConOrigArgTys destructConstructor) - 1]]

    simpleBottom b
      | b == initialBottomId         = "?"
      | otherwise                    = "?" ++ format b

    simpleOutputResult (Bottoms b) = simpleBottom b
    simpleOutputResult r           = format r
