{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE PatternGuards #-}
{-# LANGUAGE BangPatterns #-}
module IdDB.FixedTypeMap
( IdDB.FixedTypeMap.empty
, IdDB.FixedTypeMap.insert
, IdDB.FixedTypeMap.lookup
, FixedTypeMap
) where

import Control.Arrow
import DataCon
import Data.List
import Data.Monoid
import Name
import OccName
import Outputable (showPpr)
import qualified Util.UniqMap as UniqMap
import qualified Util.UniqSet as UniqSet
import TyCon
import Type
import Util.GHC
import Util (trace)
import Util.UniqMap (UniqMap)
import Util.UniqSet (UniqSet)
import Var
import VarEnv
import IdDB.TypeMap

data FixedTypeMap a  = FTM !(TypeMapNode (Last a))

empty :: FixedTypeMap a
empty = FTM Empty

lookup :: Type -> FixedTypeMap a -> Maybe a
lookup ty (FTM tm) = getLast $ realLookup ty tm 

realLookup :: (Monoid a) => Type -> TypeMapNode a -> a
{-
realLookup ty subst node
  | False <- trace ("realLookup: " ++ showDebug ty ++ ", " ++ showDebug subst) True = error "urk"
-}
realLookup _ Empty = mempty

realLookup (getTyVar_maybe -> Just t) tm
  = UniqMap.lookup (tyVars tm) t

realLookup (splitAppTy_maybe -> Just (t1,t2)) tm 
  = realLookup t2 $ realLookup t1 (appTys tm)

realLookup (splitTyConApp_maybe -> Just (tc,tys)) tm 
  = pathLookup tys $ UniqMap.lookup (tyCons tm) tc

realLookup t@(splitFunTy_maybe -> Just (t1,t2)) tm 
  = realLookup t2 $ realLookup t1 (funTys tm)

realLookup t@(splitForAllTy_maybe -> Just (tv,ty)) tm 
  = realLookup ty (UniqMap.lookup (forallTys tm) tv)

pathLookup :: Monoid a => [Type] -> TypeMapPath a -> a
pathLookup _ Stop      = mempty
pathLookup [] step     = (stopsHere step)
pathLookup (t:ts) step = pathLookup ts (realLookup t (carrysOn step))
  
insert :: FixedTypeMap a -> Type -> a -> FixedTypeMap a
insert (FTM tm) ty val = FTM (addAtType tyToAdd (Last (Just val)) tm)
  where
    tyToAdd = dropForAlls ty
