{-# LANGUAGE NamedFieldPuns #-}
{-# LANGUAGE ViewPatterns #-}
module Plan.Debug
( dotifyTreePlan
) where

import Control.Applicative hiding (empty)
import Control.Monad.State
import Data.Graph.Inductive hiding (mkNode, mkEdge)
import Data.GraphViz
import Data.Sequence (Seq)
import Format
import Plan.Experiment
import Plan.TreePlan
import qualified Data.Foldable as F
import qualified Data.Map as Map
import qualified Data.Sequence as Seq
import qualified Util.UniqMap as UniqMap
import Util
import Util.Graph
import Util.UniqMap (UniqMap)

dotifyTreePlan :: Int -> TreePlan -> String
dotifyTreePlan depth = printDotGraph . graphToDot (nonClusteredParams { fmtNode = mkNode , fmtEdge = mkEdge }) . minimizeGraph . graphTreePlan depth

mkNode :: (Format t) => (a, Maybe t) -> [Attribute]
mkNode (_, Nothing) = [Style [SItem Filled []], FillColor (RGB 0 0 0)]
mkNode (_, Just ex) = mkLabel (escapeNL $ format ex)

mkEdge :: (a,b, String) -> [Attribute]
mkEdge (_,_, x) = mkLabel x

mkLabel :: EscString -> [Attribute]
mkLabel l = [Label (StrLabel l)]

escapeNL :: String -> String
escapeNL = map (\x -> if x == '\n' then ' ' else x)

doParent :: Node -> Maybe (Node, String) -> Gr (Maybe Experiment) String 
                                         -> Gr (Maybe Experiment) String
doParent _  Nothing              = id
doParent nn (Just (preNode, er)) = insEdge (preNode, nn, er)


graphTreePlan :: Int -> TreePlan -> Gr (Maybe Experiment) String
graphTreePlan depth plan = execState ((graphTreePlan' Nothing depth plan)) empty

graphTreePlan' :: (Maybe (Node, String)) -> Int -> TreePlan -> State (Gr (Maybe Experiment) String) ()
graphTreePlan' _ 0 _ = return ()
graphTreePlan' parent n TPSplit { tpChildren }
  | Seq.null tpChildren = return ()
  | otherwise = do
    [nn] <- gets $ newNodes 1
    modify (doParent nn parent . insNode (nn, Nothing))
    F.forM_ tpChildren (graphTreePlan' (Just (nn, "")) (n-1))
graphTreePlan' parent n TPStep { tpExperiment
                               , tpOnOk
                               , tpOnDataCon
                               , tpOnBottom
                               } = do
  [nn] <- gets $ newNodes 1
  modify (doParent nn parent . insNode (nn, Just tpExperiment))
  graphTreePlan' (Just (nn, ".")) (n-1) tpOnOk
  forM_ (UniqMap.elems tpOnDataCon) $ \(dc, elt) -> do
    graphTreePlan' (Just (nn, (show dc))) (n-1) elt
  
  F.sequence_ . flip Map.mapWithKey tpOnBottom $ \botId plan -> do
    graphTreePlan' (Just (nn, "?" ++ (escapeNL $ format botId))) (n-1) plan
