module TestSuite.TestCaseBinary where

import Data.Binary
import Data.Text (Text)
import qualified Data.Text as T
import qualified Data.Text.Encoding as T
import TestSuite.TestCase

instance Binary TestSuite.TestCase.TestCase where
  put (TestCase a b) = put a >> put b
  get = get >>= \a -> get >>= \b -> return (TestCase a b)

instance Binary TestSuite.TestCase.TestResult where
  put (Errors a) = putWord8 0 >> put a
  put (Property a) = putWord8 1 >> put a
  put (Show a) = putWord8 2 >> put a
  get = do
    tag_ <- getWord8
    case tag_ of
      0 -> get >>= \a -> return (Errors a)
      1 -> get >>= \a -> return (Property a)
      2 -> get >>= \a -> return (Show a)
      _ -> fail "no parse"

instance Binary TestSuite.TestCase.TestExperiment where
  put (TestExperiment a b) = put a >> put b
  get = get >>= \a -> get >>= \b -> return (TestExperiment a b)

instance Binary TestSuite.TestCase.TestExperimentRoot where
  put (TEIdRoot a) = putWord8 0 >> put a
  put (TEDestructRoot a b c d e) = putWord8 1 >> put a >> put b >> put c >> put d >> put e
  put (TEConstantRoot a) = putWord8 2 >> put a
  get = do
    tag_ <- getWord8
    case tag_ of
      0 -> get >>= \a -> return (TEIdRoot a)
      1 -> get >>= \a -> get >>= \b -> get >>= \c -> get >>= \d -> get >>= \e -> return (TEDestructRoot a b c d e)
      2 -> get >>= \a -> return (TEConstantRoot a)
      _ -> fail "no parse"

instance Binary TestSuite.TestCase.TestExperimentConstant where
  put (TECInt a) = putWord8 0 >> put a
  put (TECChar a) = putWord8 1 >> put a
  put (TECFloat a) = putWord8 2 >> put a
  put (TECDouble a) = putWord8 3 >> put a
  put (TECInteger a) = putWord8 4 >> put a
  get = do
    tag_ <- getWord8
    case tag_ of
      0 -> get >>= \a -> return (TECInt a)
      1 -> get >>= \a -> return (TECChar a)
      2 -> get >>= \a -> return (TECFloat a)
      3 -> get >>= \a -> return (TECDouble a)
      4 -> get >>= \a -> return (TECInteger a)
      _ -> fail "no parse"

instance Binary TestSuite.TestCase.TestExperimentArg where
  put (TEArgExperiment a) = putWord8 0 >> put a
  put TEArgUnconstrained = putWord8 1
  put (TEArgBottom a) = putWord8 2 >> put a
  get = do
    tag_ <- getWord8
    case tag_ of
      0 -> get >>= \a -> return (TEArgExperiment a)
      1 -> return TEArgUnconstrained
      2 -> get >>= \a -> return (TEArgBottom a)
      _ -> fail "no parse"

instance Binary Text where
  put t = put (T.encodeUtf8 t)
  get = get >>= \a -> return (T.decodeUtf8 a)
